@@ -2,16 +2,17 @@ import re
 import os
 
 __author__ = 'Thierry Schellenbach'
-__copyright__ = 'Copyright 2014, Thierry Schellenbach'
+__copyright__ = 'Copyright 2014, Stream.io, Inc'
 __credits__ = ['Thierry Schellenbach, mellowmorning.com, @tschellenbach']
-__license__ = 'BSD'
-__version__ = '2.3.8'
+__license__ = 'BSD-3-Clause'
+__version__ = '2.3.9'
 __maintainer__ = 'Thierry Schellenbach'
-__email__ = 'thierry@getstream.io'
+__email__ = 'support@getstream.io'
 __status__ = 'Production'
 
 
-def connect(api_key=None, api_secret=None, app_id=None, version='v1.0', timeout=3.0, location=None):
+def connect(api_key=None, api_secret=None, app_id=None, version='v1.0',
+            timeout=3.0, location=None, base_url=None):
     '''
     Returns a Client object
 
@@ -32,4 +33,5 @@ def connect(api_key=None, api_secret=None, app_id=None, version='v1.0', timeout=
         else:
             raise ValueError('Invalid api key or heroku url')
 
-    return StreamClient(api_key, api_secret, app_id, version, timeout, location=location)
+    return StreamClient(api_key, api_secret, app_id, version, timeout,
+                        location=location, base_url=base_url)
