@@ -5,48 +5,73 @@ from stream.exceptions import ApiKeyException, InputException
 import random
 import jwt
 from unittest.case import TestCase
+import json
 
 import os
+import sys
 import datetime
+import datetime as dt
 import copy
 import requests
 from stream import serializer
-from requests.exceptions import ConnectionError, MissingSchema
+from requests.exceptions import MissingSchema
+from itertools import count
+from uuid import uuid4
 
 try:
     from urlparse import urlparse, parse_qs
 except ImportError:
     from urllib.parse import urlparse, parse_qs
 
+
 def connect_debug():
+    try:
+        key = os.environ['STREAM_KEY']
+        secret = os.environ['STREAM_SECRET']
+    except KeyError:
+        print('To run the tests the STREAM_KEY and STREAM_SECRET variables '
+              'need to be available. \n'
+              'Please create a pull request if you are an external '
+              'contributor, because these variables are automatically added '
+              'by Travis.')
+        sys.exit(1)
+
     return stream.connect(
-        'q56mdvdzreye',
-        'spmf6x2b2v2tqg93sfp5t393wfcxru58zm7jr3ynf7dmmndw5y8chux25hs63znf',
+        key,
+        secret,
         location='us-east',
-        timeout=10
+        timeout=30,
+        base_url='http://qa-api.getstream.io/api/',
     )
 
-random_postfix = str(int(time.time())) + str(random.randint(0, 1000))
 client = connect_debug()
 
+counter = count()
+test_identifier = uuid4()
+
+
+def get_unique_postfix():
+    return '---test_%s-feed_%s' % (test_identifier, next(counter))
+
 
 def getfeed(feed_slug, user_id):
     '''
     Adds the random postfix to the user id
     '''
-    return client.feed(feed_slug, user_id + random_postfix)
+    return client.feed(feed_slug, user_id + get_unique_postfix())
 
-user1 = getfeed('user', '1')
-user2 = getfeed('user', '2')
-aggregated2 = getfeed('aggregated', '2')
-aggregated3 = getfeed('aggregated', '3')
-topic1 = getfeed('topic', '1')
-flat3 = getfeed('flat', '3')
 
 class ClientTest(TestCase):
 
     def setUp(self):
         # DEBUG account details
+        user1 = getfeed('user', '1')
+        user2 = getfeed('user', '2')
+        aggregated2 = getfeed('aggregated', '2')
+        aggregated3 = getfeed('aggregated', '3')
+        topic1 = getfeed('topic', '1')
+        flat3 = getfeed('flat', '3')
+
         self.c = client
         self.user1 = user1
         self.user2 = user2
@@ -54,6 +79,30 @@ class ClientTest(TestCase):
         self.aggregated3 = aggregated3
         self.topic1 = topic1
         self.flat3 = flat3
+        print(user1.id)
+        print(user2.id)
+        print(aggregated2.id)
+        print(aggregated3.id)
+        print(topic1.id)
+        print(flat3.id)
+
+        self.local_tests = False
+        if 'LOCAL' in os.environ:
+            self.local_tests = os.environ['LOCAL']
+
+    def _test_sleep(self, production_wait, local_wait):
+        """
+        when testing against a live API, sometimes we need a small sleep to
+        ensure data stability, however when testing locally the wait does
+        not need to be as long
+        :param production_wait: float, number of seconds to sleep when hitting real API
+        :param local_wait: float, number of seconds to sleep when hitting localhost API
+        :return: None
+        """
+        sleep_time = production_wait
+        if self.local_tests:
+            sleep_time = local_wait
+        time.sleep(sleep_time)
 
     def test_update_activities_create(self):
         activities = [{
@@ -83,17 +132,17 @@ class ClientTest(TestCase):
                 'foreign_id': 'object:%s' % i,
                 'time': datetime.datetime.utcnow().isoformat()
             })
-        activities_created = user1.add_activities(activities)['activities']
+        activities_created = self.user1.add_activities(activities)['activities']
         activities = copy.deepcopy(activities_created)
 
-        time.sleep(3)
+
         for activity in activities:
             activity.pop('id')
             activity['popularity'] = 100
 
         self.c.update_activities(activities)
 
-        activities_updated = user1.get(limit=len(activities))['results']
+        activities_updated = self.user1.get(limit=len(activities))['results']
         activities_updated.reverse()
 
         for i, activity in enumerate(activities_updated):
@@ -116,9 +165,14 @@ class ClientTest(TestCase):
         self.assertEqual(
             client.api_secret, 'twc5ywfste5bm2ngqkzs7ukxk3pn96yweghjrxcmcrarnt3j4dqj3tucbhym5wfd')
         self.assertEqual(client.app_id, '669')
-        self.assertEqual(
-            client.base_url, 'https://api.getstream.io/api/')
-        
+
+        if self.local_tests:
+            self.assertEqual(
+                client.base_url, 'http://localhost:8000/api/')
+        else:
+            self.assertEqual(
+                client.base_url, 'https://api.getstream.io/api/')
+
     def test_heroku_location(self):
         url = 'https://ahj2ndz7gsan:gthc2t9gh7pzq52f6cky8w4r4up9dr6rju9w3fjgmkv6cdvvav2ufe5fv7e2r9qy@us-east.getstream.io/?app_id=1'
         os.environ['STREAM_URL'] = url
@@ -126,8 +180,13 @@ class ClientTest(TestCase):
         self.assertEqual(client.api_key, 'ahj2ndz7gsan')
         self.assertEqual(
             client.api_secret, 'gthc2t9gh7pzq52f6cky8w4r4up9dr6rju9w3fjgmkv6cdvvav2ufe5fv7e2r9qy')
-        self.assertEqual(
-            client.base_url, 'https://us-east-api.getstream.io/api/')
+
+        if self.local_tests:
+            self.assertEqual(
+                client.base_url, 'http://localhost:8000/api/')
+        else:
+            self.assertEqual(
+                client.base_url, 'https://us-east-api.getstream.io/api/')
         self.assertEqual(client.app_id, '1')
 
     def test_heroku_overwrite(self):
@@ -137,37 +196,47 @@ class ClientTest(TestCase):
         self.assertEqual(client.api_key, 'a')
         self.assertEqual(client.api_secret, 'b')
         self.assertEqual(client.app_id, 'c')
-        
+
     def test_location_support(self):
         client = stream.connect('a', 'b', 'c', location='us-east')
+
         full_location = 'https://us-east-api.getstream.io/api/'
+        if self.local_tests:
+            full_location = 'http://localhost:8000/api/'
+
         self.assertEqual(client.location, 'us-east')
         self.assertEqual(client.base_url, full_location)
-        # test a wrong location
-        client = stream.connect('a', 'b', 'c', location='nonexistant')
-        def get_feed():
-            client.feed('user', '1').get()
-        self.assertRaises(ConnectionError, get_feed)
-        
+
+        # test a wrong location, can only work on non-local test running
+        if not self.local_tests:
+            client = stream.connect('a',
+                                    'b',
+                                    'c',
+                                    location='nonexistant')
+            def get_feed():
+                f = client.feed('user', '1').get()
+            self.assertRaises(requests.exceptions.ConnectionError, get_feed)
+
     def test_invalid_feed_values(self):
         def invalid_feed_slug():
             client.feed('user:', '1')
         self.assertRaises(ValueError, invalid_feed_slug)
-        
+
         def invalid_user_id():
             client.feed('user:', '1-a')
         self.assertRaises(ValueError, invalid_user_id)
-            
+
         def invalid_follow_feed_slug():
             self.user1.follow('user:', '1')
         self.assertRaises(ValueError, invalid_follow_feed_slug)
-            
+
         def invalid_follow_user_id():
             self.user1.follow('user', '1-:a')
         self.assertRaises(ValueError, invalid_follow_user_id)
 
     def test_token_retrieval(self):
         self.user1.token
+        self.user1.get_readonly_token()
 
     def test_add_activity(self):
         feed = getfeed('user', 'py1')
@@ -183,7 +252,8 @@ class ClientTest(TestCase):
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
         activity_data['to'] = [team_feed.id]
         feed.add_activity(activity_data)
-        time.sleep(2)
+
+
         self.assertEqual(activity_data['to'], [team_feed.id])
 
     def test_add_activities_to_inplace_change(self):
@@ -192,7 +262,8 @@ class ClientTest(TestCase):
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
         activity_data['to'] = [team_feed.id]
         feed.add_activities([activity_data])
-        time.sleep(2)
+
+
         self.assertEqual(activity_data['to'], [team_feed.id])
 
     def test_add_activity_to(self):
@@ -206,7 +277,8 @@ class ClientTest(TestCase):
         }
         response = user_feed.add_activity(activity_data)
         activity_id = response['id']
-        time.sleep(2)
+
+
         # see if the new activity is also in the team feed
         activities = team_feed.get(limit=1)['results']
         self.assertEqual(activities[0]['id'], activity_id)
@@ -234,33 +306,44 @@ class ClientTest(TestCase):
 
         with self.assertRaises(TypeError):
             user_feed.add_activity(activity_data)
-        
+
     def assertFirstActivityIDEqual(self, activities, correct_activity_id):
         activity_id = None
         if activities:
             activity_id = activities[0]['id']
         self.assertEqual(activity_id, correct_activity_id)
-    
+
     def assertFirstActivityIDNotEqual(self, activities, correct_activity_id):
         activity_id = None
         if activities:
             activity_id = activities[0]['id']
         self.assertNotEqual(activity_id, correct_activity_id)
-        
+
     def test_remove_activity(self):
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
+
         activity_id = self.user1.add_activity(activity_data)['id']
+        activities = self.user1.get(limit=8)['results']
+        self.assertEqual(len(activities), 1)
+
         self.user1.remove_activity(activity_id)
-        activities = self.user1.get(limit=1)['results']
-        self.assertNotEqual(activities[0]['id'], activity_id)
+        # verify that no activities were returned
+        activities = self.user1.get(limit=8)['results']
+        self.assertEqual(len(activities), 0)
 
     def test_remove_activity_by_foreign_id(self):
         activity_data = {
             'actor': 1, 'verb': 'tweet', 'object': 1, 'foreign_id': 'tweet:10'}
-        activity_id = self.user1.add_activity(activity_data)['id']
+
+        self.user1.add_activity(activity_data)['id']
+        activities = self.user1.get(limit=8)['results']
+        self.assertEqual(len(activities), 1)
+
         self.user1.remove_activity(foreign_id='tweet:10')
-        activities = self.user1.get(limit=1)['results']
-        self.assertNotEqual(activities[0]['id'], activity_id)
+        # verify that no activities were returned
+        activities = self.user1.get(limit=8)['results']
+        self.assertEqual(len(activities), 0)
+
         # verify this doesnt raise an error, but fails silently
         self.user1.remove_activity(foreign_id='tweet:unknowandmissing')
 
@@ -276,7 +359,12 @@ class ClientTest(TestCase):
         self.assertEqual(get_activity_ids, activity_ids[::-1])
 
     def test_add_activities_to(self):
-        to = [getfeed('user', 'pyto2').id, getfeed('user', 'pyto3').id]
+        pyto2 = getfeed('user', 'pyto2')
+        pyto3 = getfeed('user', 'pyto3')
+        print(pyto2.id)
+        print(pyto3.id)
+
+        to = [pyto2.id, pyto3.id]
         activity_data = [
             {'actor': 1, 'verb': 'tweet', 'object': 1, 'to': to},
             {'actor': 2, 'verb': 'watch', 'object': 2, 'to': to},
@@ -287,13 +375,11 @@ class ClientTest(TestCase):
         get_activity_ids = [a['id'] for a in activities]
         self.assertEqual(get_activity_ids, activity_ids[::-1])
         # test first target
-        feed = getfeed('user', 'pyto2')
-        activities = feed.get(limit=2)['results']
+        activities = pyto2.get(limit=2)['results']
         get_activity_ids = [a['id'] for a in activities]
         self.assertEqual(get_activity_ids, activity_ids[::-1])
         # test second target
-        feed = getfeed('user', 'pyto3')
-        activities = feed.get(limit=2)['results']
+        activities = pyto3.get(limit=2)['results']
         get_activity_ids = [a['id'] for a in activities]
         self.assertEqual(get_activity_ids, activity_ids[::-1])
 
@@ -304,7 +390,8 @@ class ClientTest(TestCase):
         activity_data = {'actor': actor_id, 'verb': 'tweet', 'object': 1}
         activity_id = feed.add_activity(activity_data)['id']
         agg_feed.follow(feed.slug, feed.user_id)
-        time.sleep(10)
+
+
         activities = agg_feed.get(limit=3)['results']
         activity = self._get_first_aggregated_activity(activities)
         activity_id_found = activity['id'] if activity is not None else None
@@ -317,26 +404,29 @@ class ClientTest(TestCase):
         actor_id = random.randint(10, 100000)
         feed1.add_activity({ 'actor': actor_id, 'verb': 'tweet', 'object': 1 })
         feed.follow(feed1.slug, feed1.user_id, activity_copy_limit=0)
-        time.sleep(10)
+
+
         activities = feed.get(limit=5)['results']
 
         self.assertEqual(len(activities), 0)
-        
+
     def test_follow_and_delete(self):
         user_feed = getfeed('user', 'test_follow')
         agg_feed = getfeed('aggregated', 'test_follow')
         actor_id = random.randint(10, 100000)
         activity_data = {'actor': actor_id, 'verb': 'tweet', 'object': 1}
         activity_id = user_feed.add_activity(activity_data)['id']
+
+
         agg_feed.follow(user_feed.slug, user_feed.user_id)
         user_feed.remove_activity(activity_id)
-        time.sleep(2)
-        for x in range(5):
-            activities = agg_feed.get(limit=3)['results']
-            activity = self._get_first_aggregated_activity(activities)
-            activity_id_found = activity['id'] if activity is not None else None
-            self.assertNotEqual(activity_id_found, activity_id)
-            time.sleep(1)
+
+        activities = agg_feed.get(limit=3)['results']
+        activity = self._get_first_aggregated_activity(activities)
+        activity_id_found = (activity['id'] if activity is not None
+                             else None)
+
+        self.assertNotEqual(activity_id_found, activity_id)
 
     # def test_follow_private(self):
     #     feed = getfeed('secret', 'py1')
@@ -345,7 +435,6 @@ class ClientTest(TestCase):
     #     activity_data = {'actor': actor_id, 'verb': 'tweet', 'object': 1}
     #     activity_id = feed.add_activity(activity_data)['id']
     #     agg_feed.follow(feed.slug, feed.user_id)
-    #     time.sleep(10)
     #     activities = agg_feed.get(limit=3)['results']
     #     activity = self._get_first_aggregated_activity(activities)
     #     activity_id_found = activity['id'] if activity is not None else None
@@ -356,22 +445,28 @@ class ClientTest(TestCase):
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
         activity_id = feed.add_activity(activity_data)['id']
         self.flat3.follow(feed.slug, feed.user_id)
-        time.sleep(10)
+
+
         activities = self.flat3.get(limit=3)['results']
         activity = self._get_first_activity(activities)
         activity_id_found = activity['id'] if activity is not None else None
         self.assertEqual(activity_id_found, activity_id)
 
+        self.flat3.unfollow(feed.slug, feed.user_id)
+        activities = self.flat3.get(limit=3)['results']
+        self.assertEqual(len(activities), 0)
+
     def test_flat_follow_no_copy(self):
         feed = getfeed('user', 'test_flat_follow_no_copy')
         follower = getfeed('flat', 'test_flat_follow_no_copy')
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
         feed.add_activity(activity_data)['id']
         follower.follow(feed.slug, feed.user_id, activity_copy_limit=0)
-        time.sleep(10)
+
+
         activities = follower.get(limit=3)['results']
         self.assertEqual(activities, [])
-        
+
     def test_flat_follow_copy_one(self):
         feed = getfeed('user', 'test_flat_follow_copy_one')
         follower = getfeed('flat', 'test_flat_follow_copy_one')
@@ -380,7 +475,8 @@ class ClientTest(TestCase):
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1, 'foreign_id': 'test:2'}
         feed.add_activity(activity_data)['id']
         follower.follow(feed.slug, feed.user_id, activity_copy_limit=1)
-        time.sleep(5)
+
+
         activities = follower.get(limit=3)['results']
         # verify we get the latest activity
         self.assertEqual(activities[0]['foreign_id'], 'test:2')
@@ -480,57 +576,87 @@ class ClientTest(TestCase):
     def test_mark_read(self):
         notification_feed = getfeed('notification', 'py3')
         activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
-        notification_feed.add_activity(activity_data)
+        print(notification_feed.add_activity(activity_data)['id'])
         activity_data = {'actor': 2, 'verb': 'add', 'object': 2}
-        notification_feed.add_activity(activity_data)
+        print(notification_feed.add_activity(activity_data)['id'])
         activity_data = {'actor': 3, 'verb': 'watch', 'object': 2}
-        notification_feed.add_activity(activity_data)
+        print(notification_feed.add_activity(activity_data)['id'])
+
+
         activities = notification_feed.get(limit=3)['results']
+        from pprint import pprint
+        print(len(activities))
+        pprint(activities)
         for activity in activities:
+            pprint(activity)
             self.assertFalse(activity['is_read'])
         activities = notification_feed.get(mark_read=True)['results']
         activities = notification_feed.get(limit=2)['results']
         self.assertTrue(activities[0]['is_read'])
         self.assertTrue(activities[1]['is_read'])
-            
-    def test_mark_seen(self):
+
+    def test_get_not_marked_seen(self):
         notification_feed = getfeed('notification', 'test_mark_seen')
-        activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
-        notification_feed.add_activity(activity_data)
-        activity_data = {'actor': 2, 'verb': 'add', 'object': 2}
-        notification_feed.add_activity(activity_data)
-        activity_data = {'actor': 3, 'verb': 'watch', 'object': 3}
-        notification_feed.add_activity(activity_data)
+        print(notification_feed.add_activity({'actor': 1, 'verb': 'tweet', 'object': 1})['id'])
+        print(notification_feed.add_activity({'actor': 2, 'verb': 'tweet', 'object': 2})['id'])
+        print(notification_feed.add_activity({'actor': 3, 'verb': 'tweet', 'object': 3})['id'])
+
 
         activities = notification_feed.get(limit=3)['results']
+        from pprint import pprint
+        print(len(activities))
+        pprint(activities)
         for activity in activities:
+            pprint(activity)
             self.assertFalse(activity['is_seen'])
-            
-        activities = notification_feed.get(limit=3)['results']
-        notification_feed.get(mark_seen=[activities[0]['id'], activities[1]['id']])['results']
+
+    def test_mark_seen_on_get(self):
+        notification_feed = getfeed('notification', 'test_mark_seen')
+        activities = notification_feed.get(limit=100)['results']
+        for activity in activities:
+            notification_feed.remove_activity(activity['id'])
+
+
+        old_activities = [
+            notification_feed.add_activity({'actor': 1, 'verb': 'tweet', 'object': 1}),
+            notification_feed.add_activity({'actor': 2, 'verb': 'add', 'object': 2}),
+            notification_feed.add_activity({'actor': 3, 'verb': 'watch', 'object': 3}),
+        ]
+
+        notification_feed.get(mark_seen=[old_activities[0]['id'], old_activities[1]['id']])
+
         activities = notification_feed.get(limit=3)['results']
+
         # is the seen state correct
-        self.assertTrue(activities[0]['is_seen'])
-        self.assertTrue(activities[1]['is_seen'])
-        self.assertFalse(activities[2]['is_seen'])
+        for activity in activities:
+            # using a loop in case we're retrieving activities in a different order than old_activities
+            if old_activities[0]['id'] == activity['id']:
+                self.assertTrue(activity['is_seen'])
+            if old_activities[1]['id'] == activity['id']:
+                self.assertTrue(activity['is_seen'])
+            if old_activities[2]['id'] == activity['id']:
+                self.assertFalse(activity['is_seen'])
+
         # see if the state properly resets after we add another activity
-        activity_data = {'actor': 3, 'verb': 'watch', 'object': 3}
-        notification_feed.add_activity(activity_data)['id']
+        notification_feed.add_activity({'actor': 3, 'verb': 'watch', 'object': 3})  # ['id']
         activities = notification_feed.get(limit=3)['results']
         self.assertFalse(activities[0]['is_seen'])
         self.assertEqual(len(activities[0]['activities']), 2)
 
     def test_mark_read_by_id(self):
         notification_feed = getfeed('notification', 'py2')
-        activity_data = {'actor': 1, 'verb': 'tweet', 'object': 1}
-        notification_feed.add_activity(activity_data)['id']
-        activity_data = {'actor': 2, 'verb': 'add', 'object': 2}
-        notification_feed.add_activity(activity_data)['id']
-        activity_data = {'actor': 3, 'verb': 'watch', 'object': 2}
-        notification_feed.add_activity(activity_data)['id']
+        print(notification_feed.add_activity({'actor': 1, 'verb': 'tweet', 'object': 1})['id'])  # ['id']
+        print(notification_feed.add_activity({'actor': 2, 'verb': 'tweet', 'object': 2})['id'])  # ['id']
+        print(notification_feed.add_activity({'actor': 3, 'verb': 'tweet', 'object': 2})['id'])  # ['id']
+
+
         activities = notification_feed.get(limit=3)['results']
         ids = []
+        from pprint import pprint
+        print(len(activities))
+        pprint(activities)
         for activity in activities:
+            pprint(activity)
             ids.append(activity['id'])
             self.assertFalse(activity['is_read'])
         ids = ids[:2]
@@ -583,6 +709,8 @@ class ClientTest(TestCase):
             'actor': 1, 'verb': 'tweet', 'object': 1, 'time': utcnow}
         response = self.user1.add_activity(activity_data)
         response = self.user1.add_activity(activity_data)
+
+
         activities = self.user1.get(limit=2)['results']
         self.assertDatetimeAlmostEqual(activities[0]['time'], utcnow)
         if (len(activities) > 1):
@@ -599,7 +727,7 @@ class ClientTest(TestCase):
         self.flat3.follow('user', self.user1.user_id)
         # add the same activity twice
         now = datetime.datetime.now(tzlocal())
-        tweet = 'My Way %s' % random_postfix
+        tweet = 'My Way %s' % get_unique_postfix()
         activity_data = {
             'actor': 1, 'verb': 'tweet', 'object': 1, 'time': now, 'tweet': tweet}
         self.topic1.add_activity(activity_data)
@@ -607,24 +735,30 @@ class ClientTest(TestCase):
         # verify that flat3 contains the activity exactly once
         response = self.flat3.get(limit=3)
         activity_tweets = [a.get('tweet') for a in response['results']]
+        print(response)
+        print(activity_tweets)
         self.assertEqual(activity_tweets.count(tweet), 1)
 
     def test_uniqueness_foreign_id(self):
         now = datetime.datetime.now(tzlocal())
         utcnow = (now - now.utcoffset()).replace(tzinfo=None)
+
         activity_data = {'actor': 1, 'verb': 'tweet',
                          'object': 1, 'foreign_id': 'tweet:11', 'time': now}
         response = self.user1.add_activity(activity_data)
+
         activity_data = {'actor': 2, 'verb': 'tweet',
                          'object': 3, 'foreign_id': 'tweet:11', 'time': now}
         response = self.user1.add_activity(activity_data)
-        activities = self.user1.get(limit=2)['results']
+
+        activities = self.user1.get(limit=10)['results']
         # the second post should have overwritten the first one (because they
         # had same id)
+
+        self.assertEqual(len(activities), 1)
         self.assertEqual(activities[0]['object'], '3')
         self.assertEqual(activities[0]['foreign_id'], 'tweet:11')
         self.assertDatetimeAlmostEqual(activities[0]['time'], utcnow)
-        self.assertNotEqual(activities[1]['foreign_id'], 'tweet:11')
 
     def test_time_ordering(self):
         '''
@@ -633,13 +767,14 @@ class ClientTest(TestCase):
         add one using time
         add another activity it should be in the right spot
         '''
-        now = datetime.datetime.utcnow
+
+        # timedelta is used to "make sure" that ordering is known even though
+        # server time is not
+        custom_time = datetime.datetime.utcnow() - dt.timedelta(days=1)
+
         feed = self.user2
-        for index, activity_time in enumerate([None, now, None]):
-            time.sleep(1)
-            if activity_time is not None:
-                activity_time = activity_time()
-                middle = activity_time
+        for index, activity_time in enumerate([None, custom_time, None]):
+            self._test_sleep(1, 1) # so times are a bit different
             activity_data = {'actor': 1, 'verb': 'tweet',
                              'object': 1, 'foreign_id': 'tweet:%s' % index, 'time': activity_time}
             feed.add_activity(activity_data)
@@ -648,9 +783,9 @@ class ClientTest(TestCase):
         # the second post should have overwritten the first one (because they
         # had same id)
         self.assertEqual(activities[0]['foreign_id'], 'tweet:2')
-        self.assertEqual(activities[1]['foreign_id'], 'tweet:1')
-        self.assertEqual(activities[2]['foreign_id'], 'tweet:0')
-        self.assertDatetimeAlmostEqual(activities[1]['time'], middle)
+        self.assertEqual(activities[1]['foreign_id'], 'tweet:0')
+        self.assertEqual(activities[2]['foreign_id'], 'tweet:1')
+        self.assertDatetimeAlmostEqual(activities[2]['time'], custom_time)
 
     def test_missing_actor(self):
         activity_data = {'verb': 'tweet', 'object':
@@ -704,18 +839,37 @@ class ClientTest(TestCase):
 
     def test_follow_many_acl(self):
         sources = [getfeed('user', str(i)) for i in range(10)]
-        targets = [getfeed('flat', str(i)) for i in range(10)]
-        sources_id = [source.id for source in sources]
-        targets_id = [target.id for target in targets]
-        feeds = [{'source': s, 'target': t} for s,t in zip(sources_id, targets_id)]
+        # ensure every source is empty first
+        for feed in sources:
+            activities = feed.get(limit=100)['results']
+            for activity in activities:
+                feed.remove_activity(activity['id'])
 
-        for target in targets:
-            target.add_activity({ 'actor': 'barry', 'object': '09', 'verb': 'tweet' })
+        targets = [getfeed('flat', str(i)) for i in range(10)]
+        # ensure every source is empty first
+        for feed in targets:
+            activities = feed.get(limit=100)['results']
+            for activity in activities:
+                feed.remove_activity(activity['id'])
+        # add activity to each target feed
+        activity = {
+            'actor': 'barry',
+            'object': '09',
+            'verb': 'tweet',
+            'time': datetime.datetime.utcnow().isoformat()
+        }
+        for feed in targets:
+            feed.add_activity(activity)
+            self.assertEqual(len(feed.get(limit=5)['results']), 1)
+
+        sources_id = [feed.id for feed in sources]
+        targets_id = [target.id for target in targets]
+        feeds = [{'source': s, 'target': t} for s, t in zip(sources_id, targets_id)]
 
         self.c.follow_many(feeds, activity_copy_limit=0)
 
-        for source in sources:
-            activities = source.get(limit=5)['results']
+        for feed in sources:
+            activities = feed.get(limit=5)['results']
             self.assertEqual(len(activities), 0)
 
     def test_add_to_many(self):
@@ -723,23 +877,35 @@ class ClientTest(TestCase):
         feeds = [getfeed('flat', str(i)).id for i in range(10, 20)]
         self.c.add_to_many(activity, feeds)
 
+
         for feed in feeds:
             feed = self.c.feed(*feed.split(':'))
             self.assertEqual(feed.get()['results'][0]['custom'], 'data')
-            
+
     def test_create_email_redirect(self):
-        expected_parts = ['https://analytics.getstream.io/analytics/redirect/',
-            'auth_type=jwt',
-            'url=http%3A%2F%2Fgoogle.com%2F%3Fa%3Db%26c%3Dd',
-            'api_key=%s' % self.c.api_key,
-            'events=%5B%7B%22foreign_ids%22%3A+%5B%22tweet%3A1%22%2C+%22tweet%3A2%22%2C+%22tweet%3A3%22%2C+%22tweet%3A4%22%2C+%22tweet%3A5%22%5D%2C+%22feed_id%22%3A+%22user%3Aglobal%22%2C+%22user_id%22%3A+%22tommaso%22%2C+%22location%22%3A+%22email%22%7D%2C+%7B%22user_id%22%3A+%22tommaso%22%2C+%22label%22%3A+%22click%22%2C+%22feed_id%22%3A+%22user%3Aglobal%22%2C+%22location%22%3A+%22email%22%2C+%22position%22%3A+3%2C+%22foreign_id%22%3A+%22tweet%3A1%22%7D%5D',
-        ]
-        engagement = {'foreign_id': 'tweet:1', 'label': 'click', 'position': 3, 'user_id': 'tommaso', 'location': 'email', 'feed_id': 'user:global'}
-        impression = {'foreign_ids': ['tweet:1', 'tweet:2', 'tweet:3', 'tweet:4', 'tweet:5'], 'user_id':
-                      'tommaso', 'location': 'email', 'feed_id': 'user:global'}
-        events = [impression, engagement]
         target_url = 'http://google.com/?a=b&c=d'
         user_id = 'tommaso'
+
+        impression = {
+            'foreign_ids': ['tweet:1',
+                            'tweet:2',
+                            'tweet:3',
+                            'tweet:4',
+                            'tweet:5'],
+            'feed_id': 'user:global',
+            'user_id': user_id,
+            'location': 'email'
+        }
+        engagement = {
+            'user_id': user_id,
+            'label': 'click',
+            'feed_id': 'user:global',
+            'location': 'email',
+            'position': 3,
+            'foreign_id': 'tweet:1'
+        }
+        events = [impression, engagement]
+
         redirect_url = self.c.create_redirect_url(target_url, user_id, events)
 
         parsed_url = urlparse(redirect_url)
@@ -753,10 +919,19 @@ class ClientTest(TestCase):
             'user_id': 'tommaso'
         })
 
-        for part in expected_parts:
-            if part not in redirect_url:
-                raise ValueError('didnt find %s in url \n %s' % (part, redirect_url))
-            
+        expected_params = {
+            'auth_type': 'jwt',
+            'url': target_url,
+            'api_key': self.c.api_key,
+        }
+
+        for k, v in expected_params.items():
+            self.assertEqual(qs[k][0], v)
+
+        self.assertEqual(json.loads(qs['events'][0]), events)
+
+
+
     def test_email_redirect_invalid_target(self):
         engagement = {'foreign_id': 'tweet:1', 'label': 'click', 'position': 3, 'user_id': 'tommaso', 'location': 'email', 'feed_id': 'user:global'}
         impression = {'foreign_ids': ['tweet:1', 'tweet:2', 'tweet:3', 'tweet:4', 'tweet:5'], 'user_id':
@@ -777,6 +952,3 @@ class ClientTest(TestCase):
         res = requests.get(redirect_url)
         res.raise_for_status()
         self.assertTrue('google' in res.url)
-        
-        
-        
\ No newline at end of file
